"""Call the retriever module and implement the retrieve function. 调用retriever模块并实现检索功能

This module builds different classes according to the database oriented, and each class calls the corresponding
retriever to realize the retrieve by itself. The protein/peptide retriever is used to implement the protein/peptide
level search, and the protein-to-project mapping is implemented by the project retriever. In addition, each class
also needs to implement other functions: 1. a filtering function to remove incorrect mappings; 2. an export function
to structure the retrieve results for downstream modules to call.
此模块根据面向数据库的不同构建不同的类，每个类自行调用对应的retriever实现搜索。通过protein/peptide retriever实现蛋白质/肽段水平的检索，
在通过project retriever实现蛋白质对项目的映射。此外，每个类也需要实现其他功能：1.过滤功能，以去除错误的映射；
2.导出功能，结构化检索结果，以供下游模块调用。

    Typical usage example:
    典型用法示例：

    from patpat import hub
    from patpat import mapper
    from patpat import utility

    utility.init()
    utility.initiate_uniprot_proteome_catalog()

    identifier = 'E9PV96'
    q = hub.QueryHub()
    q.identifier = identifier
    q.simple_query()

    conf = q.get_query_config()

    m = mapper.PrideMapper()

    # Add configs
    m._identifier = conf['identifier']
    m._peptides = conf['peptides']
    m._organism = conf['organism']

    m.mapping()
    m.filtering()

    output = m.export()
"""
import time
import re
import logging

import tqdm
import requests

from . import retriever
from . import utility


class Mapper(object):
    """Base class for implementing functions such as retrieve and mapping. 用于实现检索和映射等功能的基类。

    This base class gives preliminary interface definitions to build base classes against databases.
    这个基类给出了初步的接口定义，以便针对数据库建立基类。
    """

    def mapping(self, peptide_tmp=None):
        """Functions that implement the retrieve and complete the mapping. 实现搜索和完成映射的函数。

        Parameters can be set according to actual needs.
        可以根据实际需要设置参数。
        """
        raise NotImplementedError

    def filtering(self):
        """Implementing filtering functions for incorrect mappings. 实现对不正确映射的过滤功能。"""
        raise NotImplementedError

    def export(self):
        """Implement export function for downstream module call. 为下游模块的调用实现导出功能。"""
        raise NotImplementedError

    @staticmethod
    def retriever_running(any_retriever, payload: dict = None, overtime: int = 5, sleep: int = 3):
        """Decorator or Closures. Pack the retriever to prevent processes from being interrupted
        during network fluctuations.
        装饰器或闭包。封装检索器，防止因网络波动而中断进程。

        Args:
            any_retriever: classes in retriever.py retriever模块中的类
            payload: parameters of class retriever retriever类的参数
            overtime: The number of retrieve timeouts, the default is to skip the retrieve after five errors.
                    检索超时的次数，默认是在出现五个错误后跳过检索。
            sleep: time.sleep(sleep: int)

        Returns:
            any_retriever: retriever class after retrieve, retrieve may or may not succeed.
                    经检索之后的retriever类，但检索可能成功也可能失败。
            flag: bool. False means retrieve succeed. 布尔数。False意味着检索成功。
        """
        if payload is None:
            pass
        else:
            for k, v in payload.items():
                any_retriever.payload[k] = v

        t = 0
        flag = False
        while t <= overtime:
            t += 1
            try:
                any_retriever.retrieve()
            except requests.exceptions.RequestException as e:
                logging.getLogger('core').warning(f'Error:{e}')
                time.sleep(sleep)

                if t > overtime:
                    flag = True
                    logging.getLogger('core').info(f'{t}/{overtime} times, Stop!')
                else:
                    logging.getLogger('core').info(f'{t}/{overtime} times, Try again!')

            else:
                break

        return any_retriever, flag


class PrideMapper(Mapper):
    """Mapper subclass for PRIDE database. 针对PRIDE数据库的Mapper子类。"""

    def __init__(self, payload=None):
        self.__str__ = 'PRIDE'  # name of database
        self._identifier = None  # str, Protein accession
        self._peptides = None  # list,  Peptide sequences to be retrieved
        self._organism = None  # dict, Organism to which the protein belongs. E.g.: {'name': 'xx', 'accession': 'xx'}
        self._protein_level = None  # list, retrieve results for protein levels
        self._peptides_level = None  # list, retrieve results for peptide levels
        self._project_base = None  # dict, retrieved items
        self.payload = payload
        self.overtime = 5

    def mapping(self, peptide_tmp: list = None):
        """Functions that implement the retrieve and complete the mapping. 实现检索和完成映射的函数。"""
        logging.getLogger('core').info('Select PRIDE to search.')

        # Conduct protein & peptide retrieve
        peptides_level = self.mapping2peptides()
        protein_level = self.mapping2protein()

        # Load temporary files 实现断点续传
        if peptide_tmp:
            peptides_level.extend(peptide_tmp)

        # Mapping to project
        self.mapping2project(protein_level, peptides_level)

    def mapping2protein(self):
        """Implementing protein level retrieve. 实现蛋白质水平的检索。"""
        if self._identifier is None:
            protein_level = None
        else:
            protein_retriever = retriever.PrideProteinRetriever()
            protein_retriever.request_word = self._identifier
            protein_retriever.get_payloads_on_web()

            protein_retriever, _ = self.retriever_running(protein_retriever, payload=self.payload)

            protein_level = self.protein_response_parse(protein_retriever)

        protein_level = [i for i in protein_level if self.pride_usi_detect(i[1])]
        self._protein_level = protein_level
        return protein_level

    @staticmethod
    def protein_response_parse(protein_retriever):
        """Implementing structured results of protein level retrieve. 实现蛋白质水平检索的结果结构化。"""
        t = [p['peptideevidences'] for p in protein_retriever.response.values()]
        ans_list = []
        for i in t:
            for j in i:
                ans_list.append([j['peptideSequence'], j['accession']])

        return ans_list

    def mapping2peptides(self):
        """Implementing peptide level retrieve. 实现肽段水平的检索。"""
        if self._peptides is None:
            peptides_level = None
            logging.getLogger('core').warning('No peptide can be retrieve!')
        elif isinstance(self._peptides, list):
            peptides_level = []
            logging.getLogger('core').info(f'Find {len(self._peptides)} peptides all.')
        else:
            raise TypeError

        for peptide in self._peptides:
            peptide_retriever = retriever.PridePeptideRetriever()
            peptide_retriever.request_word = peptide
            peptide_retriever.get_payloads_on_web()

            peptide_retriever, pass_flag = self.retriever_running(peptide_retriever, payload=self.payload)

            ans_list = self.peptide_response_parse(peptide_retriever)
            if ans_list:
                peptides_level.extend(ans_list)
                for i in ans_list:
                    logging.getLogger('tmp').info(f'PRIDE_peptide\t{i[0]}\t{i[1]}')
            else:
                logging.getLogger('tmp').info(f'PRIDE_peptide\t{peptide}\tNone')  # 把空白搜索加入缓存中以避免重复搜索

            if pass_flag:
                continue

        peptides_level = [i for i in peptides_level if utility.usi_detect(i[1])]
        self._peptides_level = peptides_level
        return peptides_level

    @staticmethod
    def peptide_response_parse(peptide_retriever):
        """Implementing structured results of peptide level retrieve. 实现肽段水平检索的结果结构化。"""
        t = [p['spectraevidences'] for p in peptide_retriever.response.values()]
        ans_list = []
        for i in t:
            for j in i:
                ans_list.append([j.get('peptideSequence'), j.get('usi')])
        return ans_list

    def mapping2project(self, protein_level, peptides_level):
        """Implementing mapping between retrieve result and project. 实现检索结果和项目之间的映射。"""
        logging.getLogger('core').info('Mapping to project...')
        project_base = dict()

        if protein_level:
            protein_level = [[p[0], p[1]] + [i for i in self.pride_usi_split(p[1]).values()] for p in protein_level]
            self._protein_level = protein_level
        if peptides_level:
            peptides_level = [[p[0], p[1]] + [i for i in utility.usi_split(p[1]).values()] for p in peptides_level]
            self._peptides_level = peptides_level

        projects = set([i[2] for i in peptides_level]).union(set([i[2] for i in protein_level]))

        for project in tqdm.tqdm(projects):
            project_retriever = retriever.PrideProjectRetriever()
            project_retriever.request_word = project

            project_retriever, pass_flag = self.retriever_running(project_retriever, payload=self.payload)

            project_base[project] = project_retriever.response

            if pass_flag:
                continue

        self._project_base = project_base
        logging.getLogger('core').info('PRIDE mapping completed.')

    def filtering(self):
        """Implementing remove for incorrect mappings. 排除不正确的映射。

        Remove Criteria: remove of peptide level mapping that are not from the same species.
        排除标准：排除不是同一物种的肽段水平映射。

        Returns:
                None
        Updates:
            self._project_base: dict, key: project accession, value: project content

        """
        del_projects = []
        for p in self._project_base.values():
            project_org = [i['accession'] for i in p['organisms']]
            if not self._organism['accession'] in project_org:
                del_projects.extend([p['accession']])
        for p in del_projects:
            self._project_base.pop(p)

    def export(self):
        """Implement export function for downstream module call. 为下游模块的调用实现导出功能。

        Returns:
            projects: dict. Structured project information. 结构化的项目信息。
        """
        projects = self._project_base.copy()
        for project in projects:
            if self._peptides_level:
                peptides = [i[1] for i in self._peptides_level if i[2] == project]
            else:
                peptides = None
            if self._protein_level:
                protein = [i[1] for i in self._protein_level if i[2] == project]
            else:
                protein = None

            # These five keys are necessary 这五个key是必要的

            projects[project]['title'] = projects[project]['title']

            # Used to record the mapping of protein and peptide levels
            projects[project]['peptides'] = peptides
            projects[project]['protein'] = protein

            # Used to record the summary of project
            projects[project]['summary'] = projects[project]['projectDescription']

            # Used to record the website of project
            projects[project]['website'] = f'http://proteomecentral.proteomexchange.org/cgi/GetDataset?ID={project}'

        return projects

    @staticmethod
    def pride_usi_split(usi):
        """Splitting PRIDE USI"""
        component = dict()

        collection = re.search('\\S*?(?=:)', usi)

        component['collection'] = collection.group()

        split_point1 = collection.end() + 1
        split_point2 = re.search('[OPQ]\\d[A-Z\\d]{3}\\d|[A-NR-Z]\\d([A-Z][A-Z\\d]{2}\\d){1,2}:\\d*', usi).end()

        component['msRun+index'] = usi[split_point1:split_point2 - 1]
        component['interpretation'] = usi[split_point2:]

        return component

    @staticmethod
    def pride_usi_detect(usi):
        """Detect if the string is PRIDE USI"""
        try:
            c = PrideMapper.pride_usi_split(usi)
        except AttributeError:
            return False

        else:
            if re.search("pxd", c['collection'], re.I) and c['msRun+index'] and c['interpretation']:
                return True
            else:
                return False


class IProXMapper(Mapper):
    """Mapper subclass for iProX database. 针对iProX数据库的Mapper子类。"""

    def __init__(self, payload=None):
        self.__str__ = 'iProX'  # name of database
        self._identifier = None  # str, Protein accession
        self._peptides = None  # list,  Peptide sequences to be retrieved
        self._organism = None  # dict, Organism to which the protein belongs. E.g.: {'name': 'xx', 'accession': 'xx'}
        self._protein_level = None  # list, retrieve results for protein levels
        self._peptides_level = None  # list, retrieve results for peptide levels
        self._project_base = None  # dict, retrieved items
        self.payload = payload
        self.overtime = 5

    def mapping(self, peptide_tmp: list = None):
        """Functions that implement the retrieve and complete the mapping. 实现检索和完成映射的函数。"""
        logging.getLogger('core').info('Select iProX to search.')

        # Conduct protein & peptide retrieve
        peptides_level = self.mapping2peptides()
        protein_level = self.mapping2protein()

        # Load temporary files
        if peptide_tmp:
            peptides_level.extend(peptide_tmp)

        # Mapping to project
        self.mapping2project(protein_level, peptides_level)

    def mapping2protein(self):
        """Implementing protein level retrieve. 实现蛋白质水平的检索。"""
        if self._identifier is None:
            protein_level = None
        else:
            protein_retriever = retriever.IProXProteinRetriever()
            protein_retriever.request_word = self._identifier
            protein_retriever.get_payloads_on_web()

            protein_retriever, _ = self.retriever_running(protein_retriever, payload=self.payload)

            protein_level = self.protein_response_parse(protein_retriever)

        protein_level = [i for i in protein_level if utility.usi_detect(i[1])]
        self._protein_level = protein_level
        return protein_level

    @staticmethod
    def protein_response_parse(protein_retriever):
        """Implementing structured results of protein level retrieve. 实现蛋白质水平检索的结果结构化。"""
        ans_list = []

        for ps in protein_retriever.response.values():
            ans_list.extend([[p['peptideSequence'], p['usi']] for p in ps])

        return ans_list

    def mapping2peptides(self):
        """Implementing peptide level retrieve. 实现肽段水平的检索。"""
        if self._peptides is None:
            peptides_level = None
            logging.getLogger('core').warning('No peptide can be retrieve!')
        elif isinstance(self._peptides, list):
            peptides_level = []
            logging.getLogger('core').info(f'Find {len(self._peptides)} peptides all.')
        else:
            raise TypeError

        if self._peptides:
            for n, peptide in enumerate(self._peptides):
                peptide_retriever = retriever.IProXPeptideRetriever()
                peptide_retriever.request_word = peptide
                peptide_retriever.get_payloads_on_web()

                peptide_retriever, pass_flag = self.retriever_running(peptide_retriever, payload=self.payload)

                time.sleep(n / len(self._peptides))  # Take a break

                ans_list = self.peptide_response_parse(peptide_retriever)
                if ans_list:
                    peptides_level.extend(ans_list)

                    for i in ans_list:
                        logging.getLogger('tmp').info(f'iProX_peptide\t{i[0]}\t{i[1]}')
                else:
                    logging.getLogger('tmp').info(f'iProX_peptide\t{peptide}\tNone')
                if pass_flag:
                    continue

        peptides_level = [i for i in peptides_level if utility.usi_detect(i[1])]
        self._peptides_level = peptides_level
        return peptides_level

    @staticmethod
    def peptide_response_parse(peptide_retriever):
        """Implementing structured results of peptide level retrieve. 实现肽段水平检索的结果结构化。"""
        ans_list = []
        requests_word = peptide_retriever.request_word
        for ps in peptide_retriever.response.values():
            for p in ps:
                interpretation = utility.usi_split(p['usi'])['interpretation']
                answer_word = re.search('.*(?=/)', re.sub('\\[.*?]', '', interpretation)).group()
                if requests_word == answer_word:
                    ans_list.append([requests_word, p['usi']])

        return ans_list

    def mapping2project(self, protein_level, peptides_level):
        """Implementing mapping between retrieve result and project. 实现检索结果和项目之间的映射。"""
        logging.getLogger('core').info('Mapping to project...')
        project_base = dict()

        if protein_level:
            protein_level = [[p[0], p[1]] + [i for i in utility.usi_split(p[1]).values()] for p in protein_level]
            self._protein_level = protein_level
        if peptides_level:
            peptides_level = [[p[0], p[1]] + [i for i in utility.usi_split(p[1]).values()] for p in peptides_level]
            self._peptides_level = peptides_level

        projects = set([i[2] for i in peptides_level]).union(set([i[2] for i in protein_level]))

        for project in tqdm.tqdm(projects):
            project_retriever = retriever.IProXProjectRetriever()
            project_retriever.request_word = project

            project_retriever, pass_flag = self.retriever_running(project_retriever, payload=self.payload)

            project_base[project] = project_retriever.response

            if pass_flag:
                continue

        self._project_base = project_base
        logging.getLogger('core').info('iProX mapping completed.')

    def filtering(self):
        """Implementing remove for incorrect mappings. 排除不正确的映射。

        Remove Criteria: remove of peptide level mapping that are not from the same species.
        排除标准：排除不是同一物种的肽段水平映射。

        Returns:
            None

        Updates:
            Update self.project_base: dict, key: project accession, value: project content
        """
        projects = [
            p['accession']['value'] for p in self._project_base.values()
            if self._organism['accession'] in [
                i['name'] for i in p['species']
                if i['accession'] == 'MS:1001467'
            ]
        ]

        project_base = dict()
        for p in projects:
            project_base[p] = self._project_base[p]
        self._project_base = project_base

    def export(self):
        """Implement export function for downstream module call. 为下游模块的调用实现导出功能。

        Returns:
            projects: dict. Structured project information. 结构化的项目信息。
        """
        projects = self._project_base.copy()
        for project in projects:
            if self._peptides_level:
                peptides = [i[1] for i in self._peptides_level if i[2] == project]
            else:
                peptides = None
            if self._protein_level:
                protein = [i[1] for i in self._protein_level if i[2] == project]
            else:
                protein = None

            # These five keys are necessary 这五个key是必要的

            projects[project]['title'] = projects[project]['title']

            # Used to record the mapping of protein and peptide levels
            projects[project]['peptides'] = peptides
            projects[project]['protein'] = protein

            # Used to record the summary of project
            projects[project]['summary'] = projects[project]['summary']

            # Used to record the website of project
            projects[project]['website'] = f'http://proteomecentral.proteomexchange.org/cgi/GetDataset?ID={project}'

        return projects


class MassIVEMapper(Mapper):
    """Mapper subclass for MassIVE database. 针对 MassIVE 数据库的 Mapper 子类。"""

    def __init__(self, payload=None):
        self.__str__ = 'MassIVE'  # name of database
        self._identifier = None  # str, Protein accession
        self._peptides = None  # list,  Peptide sequences to be retrieved
        self._organism = None  # dict, Organism to which the protein belongs. E.g.: {'name': 'xx', 'accession': 'xx'}
        self._protein_level = None  # list, retrieve results for protein levels
        self._peptides_level = None  # list, retrieve results for peptide levels
        self._project_base = None  # dict, retrieved items
        self.payload = payload
        self.overtime = 5

    def mapping(self, peptide_tmp=None):
        """Functions that implement the retrieve and complete the mapping. 实现检索和完成映射的函数。"""
        logging.getLogger('core').info('Select MassIVE to search.')

        # Conduct protein & peptide retrieve
        peptides_level = self.mapping2peptides()
        protein_level = self.mapping2protein()

        # Mapping to project
        self.mapping2project(protein_level, peptides_level)

    def mapping2protein(self):
        """Implementing protein level retrieve. 实现蛋白质水平的检索。"""
        if not self._identifier:
            protein_level = None
        else:
            protein_retriever = retriever.MassIVEPro2ProjectRetriever()
            protein_retriever.request_word = self._identifier
            test_request = protein_retriever.get_payloads_on_web()

            if not test_request.ok:
                logging.getLogger('core').info(f'Protein {self._identifier} no request.')

            protein_retriever, _ = self.retriever_running(protein_retriever, payload=self.payload)

            protein_level = self.protein_response_parse(protein_retriever)

        self._protein_level = protein_level
        return protein_level

    @staticmethod
    def protein_response_parse(protein_retriever):
        """Implementing structured results of protein level retrieve. 实现蛋白质水平检索的结果结构化。"""
        ans_list = []

        if protein_retriever.response:
            for project in protein_retriever.response.values():
                ans_list.extend(project)

            for project in ans_list:
                project['target'] = {"protein": [protein_retriever.request_word]}

        return ans_list

    def mapping2peptides(self):
        """Implementing peptide level retrieve. 实现肽段水平的检索。"""
        if self._peptides is None:
            peptides_level = None
            logging.getLogger('core').warning('No peptide can be retrieve!')
        elif isinstance(self._peptides, list):
            peptides_level = []
            logging.getLogger('core').info(f'Find {len(self._peptides)} peptides all.')
        else:
            raise TypeError

        if self._peptides:
            for n, peptide in enumerate(self._peptides):
                peptide_retriever = retriever.MassIVEPep2ProjectRetriever()
                peptide_retriever.request_word = peptide
                test_request = peptide_retriever.get_payloads_on_web()

                if not test_request:
                    logging.getLogger('core').info(f'Peptide {peptide} no request.')
                    continue

                peptide_retriever, pass_flag = self.retriever_running(peptide_retriever, payload=self.payload)

                # time.sleep(n / len(self._peptides))  # Take a break

                ans_list = self.peptide_response_parse(peptide_retriever)

                if ans_list:
                    peptides_level.extend(ans_list)

                if pass_flag:
                    continue

        self._peptides_level = peptides_level
        return peptides_level

    @staticmethod
    def peptide_response_parse(peptide_retriever):
        """Implementing structured results of protein level retrieve. 实现蛋白质水平检索的结果结构化。"""
        ans_list = []

        if peptide_retriever.response:
            for project in peptide_retriever.response.values():
                ans_list.extend(project)

            for project in ans_list:
                project['target'] = {"peptides": [peptide_retriever.request_word]}

        return ans_list

    def mapping2project(self, protein_level, peptides_level):
        """Implementing mapping between retrieve result and project. 实现检索结果和项目之间的映射。"""
        logging.getLogger('core').info('Mapping to project...')
        project_base = dict()

        if protein_level:
            for project in protein_level:
                try:
                    project_base[f"{project['dataset_name']}"]
                except KeyError:
                    project_base[f"{project['dataset_name']}"] = {'MASSIVE': project}
                else:
                    try:
                        project_base[f"{project['dataset_name']}"]['MASSIVE']['target']['protein']
                    except KeyError:
                        project_base[f"{project['dataset_name']}"]['MASSIVE']['target']['protein'] = \
                            project['target']["protein"]
                    else:
                        project_base[f"{project['dataset_name']}"]['MASSIVE']['target']['protein'].extend(
                            project['target']["protein"])

        if peptides_level:
            for project in peptides_level:
                try:
                    project_base[f"{project['dataset_name']}"]
                except KeyError:
                    project_base[f"{project['dataset_name']}"] = {'MASSIVE': project}
                else:
                    try:
                        project_base[f"{project['dataset_name']}"]['MASSIVE']['target']['peptides']
                    except KeyError:
                        project_base[f"{project['dataset_name']}"]['MASSIVE']['target']['peptides'] = \
                            project['target']["peptides"]
                    else:
                        project_base[f"{project['dataset_name']}"]['MASSIVE']['target']['peptides'].extend(
                            project['target']["peptides"])

        for project in tqdm.tqdm(project_base.keys()):
            project_retriever = retriever.MassIVEProjectRetriever()
            project_retriever.request_word = project

            project_retriever, pass_flag = self.retriever_running(project_retriever, payload=self.payload)

            project_base[project]['PROXI'] = project_retriever.response

            if pass_flag:
                continue

        t = [i['MASSIVE']['dataset_name'] for i in project_base.values() if not i['PROXI']]
        for i in t:
            project_base.pop(i)

        self._project_base = project_base
        logging.getLogger('core').info('MassIVE mapping completed.')

    def filtering(self):
        """Implementing remove for incorrect mappings. 排除不正确的映射。

        Remove Criteria: remove of peptide level mapping that are not from the same species.
        排除标准：排除不是同一物种的肽段水平映射。

        Returns:
            project_base: dict, filtered projects

        Updates:
            Update self.project_base: dict, key: project accession, value: project content
        """

        all_project = set(self._project_base.keys())
        pass_project = set()
        for project in self._project_base.values():
            for species in project['PROXI']['species']:
                t = []
                for s in species:
                    if s['accession'] == 'MS:1001467':
                        t.extend([s['value']])
                if self._organism['accession'] in t:
                    pass_project.add(project['MASSIVE']['dataset_name'])

        del_project = all_project - pass_project

        project_base = self._project_base.copy()
        for p in del_project:
            project_base.pop(p)

        self._project_base = project_base
        return project_base

    def export(self):
        """Implement export function for downstream module call. 为下游模块的调用实现导出功能。

        Returns:
            projects: dict. Structured project information. 结构化的项目信息。
        """
        projects = self._project_base.copy()

        for project in projects.keys():

            # These five keys are necessary 这五个key是必要的

            # Used to record the title of project
            projects[project]['title'] = projects[project]['PROXI']['title']

            # Used to record the mapping of protein and peptide levels
            projects[project]['peptides'] = None
            projects[project]['protein'] = None
            # Used to record the summary of project
            projects[project]['summary'] = projects[project]['MASSIVE']['description']

            # Used to record the website of project
            projects[project]['website'] = f'https://massive.ucsd.edu/ProteoSAFe/FindDatasets?query={project}'

        return projects



"""This module provides users with a one-stop shop for using patpat_env classes. 本模块为用户提供一站式使用patpat功能的类。

This module is the core module of patpat_env and allows users to aggregate other classes of patpat_env through the classes
in this module. init function provides the runtime environment required for patpat_env. queryHub class provides querying
of protein metadata and generating peptides to be retrieved, while MapperHub class provides breakpoints for merging
and retrieving multiple proteomic databases . Furthermore, both classes are pluggable, so it is easy to insert
self-built methods or classes into them, as long as the developer adheres to the interface design.
本模块是patpat的核心模块，用户可以通过本模块中的类将patpat的其他类聚合在一起使用。
QueryHub类提供对蛋白质元数据的查询和需要检索肽段的生成功能，MapperHub类则提供多蛋白质组数据库合并检索和检索断点续传的功能。
其次，这两个类都是可拔插的，因此只要开发者遵守接口设计，很容易将自建的方法或类插入其中。

    Typical usage example:
    典型用法示例：

    from patpat import hub
    from patpat import mapper
    from patpat import utility

    utility.init()
    utility.initiate_uniprot_proteome_catalog()

    identifier_ = 'P05067'

    q = hub.QueryHub()
    q.identifier = identifier_
    q.simple_query()

    conf_ = q.get_query_config()
    mappers_ = [mapper.PrideMapper(), mapper.IProXMapper()]

    m = hub.MapperHub(config=conf_,
                      mappers=mappers_,
                      task=None
                      # task=[your task's uuid]
                      )

    m.mapping()
    result_ = m.export()
"""
import json
import os
import re
import sys
import time
import logging
import uuid

from . import utility
from . import querier
from . import mapper
from . import logger
from . import checker


class QueryHub(object):
    """This class is used to build the configuration information that needs to be retrieved.
       本类用于建立需要检索的配置信息。

    This class uses the UniProt database by default to query protein metadata and generate the peptides to
    be retrieved ,using a local library build method, the process is encapsulated in the simple_query function.
    The developer encourages users to experiment with the pluggable QueryHub class to match specific search
    requirements or to insert new peptide generation methods.
    本类默认使用UniProt数据库查询蛋白质元数据，使用本地建库的方法生成需要检索肽段，这个过程被封装在simple_query函数中。
    开发者鼓励用户尝试使用可拔插的QueryHub类以符合特殊的搜索要求，或插入新的肽段生成方法。
    """

    def __init__(self):
        self.identifier: str = ''  # protein sequence
        self.peptides: list = []  # Filtered peptides for output to external
        self.organism: dict = dict()  # dict, Organism to which the protein belongs.

        self.fasta: dict = dict()  # Structuring protein fasta into a dictionary.
        self.source: str = ''  # Local proteome file address

        self.digestion_params: list = []  # Parameters of protein in-silico enzymatic

        # pluggable
        self._protein_querier = None  # class ProteinQuerier
        self._peptide_querier = None  # class PeptideQuerier

    @property
    def protein_querier(self):
        return self._protein_querier

    @protein_querier.setter
    def protein_querier(self, protein_querier):
        """Set class ProteinQuerier"""
        self._protein_querier = protein_querier

    @property
    def peptide_querier(self):
        return self._peptide_querier

    @peptide_querier.setter
    def peptide_querier(self, peptide_querier):
        """Set class PeptideQuerier"""
        self._peptide_querier = peptide_querier

    def protein_query(self):
        """Call the class ProteinQuerier in itself. 在自身中调用ProteinQuerier类。

        Call the query method and get_properties method of class ProteinQuerier to query protein metadata and export it.
        调用ProteinQuerier类的query方法和get_properties方法，查询蛋白质元数据并导出。

        Returns:
            self.fasta['sequence']: str
            self.organism: dict

        Updates:
            self.protein_querier
            self.identifier
            self.organism
            self.fasta
        """
        self.protein_querier.query()
        self.identifier, self.organism, self.fasta = self.protein_querier.get_properties()

        return self.fasta['sequence'], self.organism

    def peptide_query(self):
        """Call the class PeptideQuerier in itself. 在自身中调用PeptideQuerier类。

        Call the query method and get_properties method of the ProteinQuerier class to generate the sequence of
        peptides to be retrieved and export them.
        调用ProteinQuerier类的query方法和get_properties方法，生成要检索的肽段序列并导出。

        Returns:
            self.digestion_params: dict
            self.source: str
            self.peptides: list

        Updates:
            self.peptide_querier
            self.digestion_params
            self.source
            self.peptides
        """
        self.peptide_querier.query()
        self.digestion_params, self.source, self.peptides = self.peptide_querier.get_properties()

        return self.digestion_params, self.source, self.peptides

    def simple_query(self):
        """Default query methods.

        This function encapsulates the process of using the UniProt database to query protein metadata and
        the local library building method to generate the peptides to be retrieved.
        本函数封装了使用UniProt数据库查询蛋白质元数据，和本地建库的方法生成需要检索肽段的过程。

        Returns:
            None

        Updates:
            self.protein_querier
            self.peptide_querier
        """
        self.protein_querier = querier.UniProtProteinQuerier()
        self.protein_querier.set_params(accession=self.identifier)
        self.protein_query()

        self.peptide_querier = querier.LocalPeptideQuerier()
        self.peptide_querier.set_params(sequence=self.fasta['sequence'],
                                        organism=self.organism)
        self.peptide_query()

    def get_query_config(self):
        """Get query config.

        Returns:
            config: dict, {'identifier': self.identifier,
                           'peptides': self.peptides,
                           'organism': self.organism,
                           'digestion': self.digestion_params,
                           'proteome_source': self.source,
                           'description': self.fasta['description']
                          }
        """
        config = {'identifier': self.identifier,
                  'peptides': self.peptides,
                  'organism': self.organism,
                  'digestion': self.digestion_params,
                  'proteome_source': self.source,
                  'description': self.fasta['description']
                  }

        return config


class MapperHub:
    """This class is used to aggregate multiple proteomic database Mapper, merge and retrieve proteins.
       本类用于聚合多个蛋白质组数据库Mapper，对蛋白质进行合并检索。

    The merge retrieve function of this class only requires the user to insert the Mapper that needs to be retrieved,
    and the developer encourages users to try to build their own Mapper. In addition to the merge search function,
    this class also provides functions such as breakpoint and logging, which users can use by setting the task property.
    本类的合并检索功能只需要用户插入需要检索的Mapper即可，开发者鼓励用户尝试自建Mapper。除了合并搜索功能，
    本类还提供断点续传和日志等功能，用户可通过设置task属性使用。
    """

    def __init__(self, config: dict, mappers: list, task=None):
        # Set working task
        if task:
            self._task = task
        else:
            self._task = str(uuid.uuid4())
        self._tmp = dict()  # temporary data

        # Load the list of classes in the Mapper module, only the classes in this list can be inserted.
        # 载入Mapper模块中的类列表，只有在该列表内的类才可插入。
        self.supported_mappers = mapper.Mapper.__subclasses__()
        self.config = None  # configuration information that needs to be retrieved 用于检索的配置信息
        self.logger = None  # class Logger
        self.mappers = set()  # class Mapper

        self._set_logging()
        self._load_tmp()
        self._set_mappers(mappers)
        self._set_mapping_config(config)
        self._set_task_info()

    def _set_logging(self):
        """Configuring Logger"""
        self.logger = logger.CoreLogger(self._task)
        self.logger.set_core()
        self.logger.set_tmp()

    def _load_tmp(self):
        """Loading temporary data, implementing breakpoint transfer.

        Insert the new Mapper if you want to implement breakpoint transfer, you need to change the function.
        插入新建的Mapper类，若想要实现断点续传，需更改该函数。
        """
        task = self._task
        if os.path.exists(f'patpat_env/tmp/{task}.log'):
            with open(f'patpat_env/tmp/{task}.log', mode='r') as f:
                data = [re.split('\t', i) for i in f.readlines()]
            self._tmp['PRIDE_peptide'] = [[i[1], i[2][:-1]] for i in data if i[0] == 'PRIDE_peptide']
            self._tmp['iProX_peptide'] = [[i[1], i[2][:-1]] for i in data if i[0] == 'iProX_peptide']
        else:
            self._tmp['PRIDE_peptide'] = []
            self._tmp['iProX_peptide'] = []

    def _set_mapping_config(self, config: dict):
        """Update configuration"""
        self.config = config.copy()
        self.config['task'] = self._task
        self.config['mappers'] = []
        for m in self.mappers:
            self.config['mappers'].extend([m.__str__])
        self.config['state'] = 'Preparation'

    def _set_mappers(self, mappers: {list, set}):
        """Set Mappers"""
        for mapper_ in mappers:
            if (type(mapper_) in self.supported_mappers and
                    not type(mapper_) in [type(m) for m in self.mappers]):
                self.mappers.add(mapper_)

    def mapping(self):
        """Core methods of this class

        Insert the new Mapper if you want to implement breakpoint transfer, you need to change the function.
        插入新建的Mapper类，若想要实现断点续传，需更改该函数。
        """
        mappers = set()
        self.config['startTime'] = time.time()
        self.config['state'] = 'Running'
        self._set_task_info()

        try:
            for mapper_ in self.mappers:

                # These three keys (protein UniProt identifier, peptide to be searched, species) are
                # required in the configuration.
                # 这三个键（蛋白UniProt识别符、需要搜索的肽段、物种）是配置中必须的。
                mapper_._identifier = self.config['identifier']
                mapper_._peptides = self.config['peptides']
                mapper_._organism = self.config['organism']

                # Implement breakpoint transfer
                tmp = None
                if mapper_.__class__ is mapper.PrideMapper and self._tmp['PRIDE_peptide']:
                    tmp = self._tmp['PRIDE_peptide']
                elif mapper_.__class__ is mapper.IProXMapper and self._tmp['iProX_peptide']:
                    tmp = self._tmp['iProX_peptide']

                if tmp:
                    tmp_peptides = set([i[0] for i in tmp])
                    mapper_._peptides = list(set(mapper_._peptides) - tmp_peptides)
                    for p in tmp_peptides:
                        logging.getLogger('core').debug(f'Local loading: {p}')

                    tmp = [i for i in tmp if utility.usi_detect(i[1])]  # 去掉空白搜索

                mapper_.mapping(peptide_tmp=tmp)
                mapper_.filtering()
                mappers.add(mapper_)
            self.mappers = mappers
        except:
            self.config['state'] = 'Error'
            logging.getLogger('core').error('RUNNING ERROR')
            self._set_task_info()
            raise BufferError(f'RUNNING ERROR:{sys.exc_info()}')
        else:
            self.config['state'] = 'Success'
            self._set_task_info()

    def _set_task_info(self):
        """Write information about this task to a local file."""
        if os.path.exists('patpat_env/logs/tasks.json'):
            with open('patpat_env/logs/tasks.json', 'r') as fr:
                configs = json.loads(fr.readline())
                config = self.config.copy()
                config['peptides'] = len(config['peptides'])
                configs['tasks'][config['task']] = config
            fr.close()
            with open('patpat_env/logs/tasks.json', 'w') as fw:
                configs_json = json.dumps(configs)
                fw.write(configs_json)
        else:
            config = self.config.copy()
            config['peptides'] = len(config['peptides'])
            configs = dict()
            configs['tasks'] = {config['task']: config}
            configs_json = json.dumps(configs)
            with open('patpat_env/logs/tasks.json', mode='w') as f:
                f.write(configs_json)

    def export(self):
        """Exporting retrieved results.

        Returns:
            output: dir
        """
        output = dict()
        for m in self.mappers:
            output[m.__str__] = m.export()

        if os.path.exists(f'patpat_env/result/{self._task}'):
            pass
        else:
            os.mkdir(f'patpat_env/result/{self._task}')

        with open(f'patpat_env/result/{self._task}/config.json', 'w') as fw:
            config = json.dumps(self.config)
            fw.write(config)

        with open(f'patpat_env/result/{self._task}/result.json', 'w') as fw:
            output_json = json.dumps(output)
            fw.write(output_json)

        with open(f'patpat_env/result/{self._task}/result.tsv', 'w') as fw:
            fw.write('title\tsummary\twebsite\n')
            for projects in output.values():
                for project in projects.values():
                    try:
                        fw.write(f"{project['title']}\t{project['summary']}\t{project['website']}\n")
                    except UnicodeEncodeError:
                        fw.write(
                            f"{project['title']}\t{project['summary'].encode(errors='ignore')}\t{project['website']}\n")

        return output


class CheckerHub:
    """Call all of Checker"""
    def __init__(self):
        self.recommended_mappers = []
        self.checkers = None

    def check(self):
        checkers = checker.GenericChecker.__subclasses__()

        self.checkers = [c() for c in checkers]

        for c in self.checkers:
            m = c.check()
            if m:
                self.recommended_mappers.extend([m])

        return self.recommended_mappers




"""This module contains patpat_env's self-built methods. 本模块包含patpat自建的方法。"""

import collections
import os
import re
import time
import json
import gzip
from urllib import parse
from ftplib import FTP

import requests
import tqdm
import pandas as pd


def init():
    """Used to create package runtime environment. 用于创建包运行环境。

    Directory structure of the runtime environment:
        patpat_env/
            |-- logs/
            |-- tmp/
            |-- result/
            |-- proteome/
    """

    def init_subdir():
        """Used to create subdirectory. 用于创建子目录"""
        dir_list = ['logs', 'tmp', 'result', 'proteome']
        for dir_ in dir_list:
            try:
                os.makedirs(f'patpat_env/{dir_}')
            except FileExistsError:
                print(f'\033[0;31mWarning! This directory already exists. 警告！该目录已经存在。[patpat_env/{dir_}]\033[0m')

    if os.path.exists('patpat_env/'):
        print(f'\033[0;31mWarning! This directory already exists. 警告！该目录已经存在。[patpat_env/]\033[0m')
        init_subdir()
    else:
        os.mkdir('patpat_env/')
        init_subdir()
    print('Runtime environment completed.')


def flatten(item, ignore_types=(str, bytes, set)):
    for x in item:
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, ignore_types):
            yield from flatten(x)
        else:
            yield x


def initiate_uniprot_proteome_catalog():
    """Initiate UniProt's proteome catalog

    Returns:
        patpat_env/
            |-- proteome/
                |-- UP_README_yyyy-mm-dd
    """

    ftp = FTP('ftp.uniprot.org')
    ftp.login()
    ftp.cwd('./pub/databases/uniprot/current_release/knowledgebase/reference_proteomes/')
    download_time = time.strftime('%Y-%m-%d', time.localtime())
    with open(f'patpat_env/proteome/UP_README_{download_time}', mode='wb') as f:
        ftp.retrbinary(f'RETR README', f.write)
    ftp.close()
    print('Uniprot_proteome_list has been initiated.')


def create_list_uniprot_proteome():
    """Create UniProt Proteome Catalog"""
    os.chdir('patpat_env/proteome/')
    file = [i for i in os.listdir() if re.search('UP_README', i)][0]
    with open(file, mode='r') as f:
        flag = ''
        while flag != 'Proteome_ID	Tax_ID	OSCODE	SUPERREGNUM	#(1)	#(2)	#(3)	Species Name\n':
            flag = f.readline()
        r = [re.split('\t', flag[:-1])]
        while True:
            flag = f.readline()
            try:
                flag[1]
            except IndexError:
                break
            else:
                r.append(re.split('\t', flag[:-1]))
        df = pd.DataFrame(r[1:], columns=r[0])
        os.chdir('../..')
    return df


def list_local_proteome():
    """List local proteome fasta"""
    proteome = os.listdir('patpat_env/proteome')
    return [re.search('(?<=uniprot-proteome_).*(?=.fasta)', p).group()
            for p in proteome if re.search('(?<=uniprot-proteome_).*(?=.fasta)', p)]


def download_uniprot_opg(taxonomy_id):
    """Download OPG Proteome fasta from UniProt"""
    df = create_list_uniprot_proteome()
    target = df.loc[df['Tax_ID'] == str(taxonomy_id)]
    tax_division = target['SUPERREGNUM'].values[0].capitalize()
    up_id = target['Proteome_ID'].values[0]

    ftp = FTP('ftp.uniprot.org')
    ftp.login()
    ftp.cwd(f'./pub/databases/uniprot/current_release/'
            f'knowledgebase/reference_proteomes/{tax_division}/{up_id}')

    file_name = up_id + '_' + target['Tax_ID'].values[0] + '.fasta'
    download_time = time.strftime('%Y-%m-%d', time.localtime())

    with open(f'patpat_env/proteome/{file_name}.gz',
              mode='wb') as f_gz:
        ftp.retrbinary(f'RETR {file_name}.gz', f_gz.write)
    ftp.close()

    gz = gzip.GzipFile(f'patpat_env/proteome/{file_name}.gz', mode='rb')
    with open(f'patpat_env/proteome/uniprot-proteome_{up_id}_OGP_{download_time}.fasta', mode='wb') as fo:
        fo.write(gz.read())
    gz.close()

    print(f'{file_name} download is complete.')
    os.remove(f'patpat_env/proteome/{file_name}.gz')
    return f'patpat_env/proteome/uniprot-proteome_{up_id}_OGP_{download_time}.fasta'


def download_uniprot_proteome(identifier: str):
    """Download proteome fasta from UniProt database

    Args:
        identifier: str.
                The 'proteome' filter value has invalid format, it should match the regular expression UP[0-9]{9}
    """
    url = f'https://rest.uniprot.org/uniprotkb/stream?format=fasta&query=%28%28proteome%3A{identifier}%29%29'

    protein_fasta = requests.get(url).text

    # FASTA name format: uniprot-proteome_UPxxxxxxxxx_yyyy-mm-dd.fasta
    download_time = time.strftime('%Y-%m-%d', time.localtime())
    fasta_name = '_'.join(['patpat_env/proteome/uniprot-proteome',
                           identifier,
                           download_time]) + '.fasta'
    with open(fasta_name, mode='w') as fo:
        fo.write(protein_fasta)
    print('Download complete.')


def pagination_download_uniprot_proteome(identifier: str, reviewed: bool = True, size: int = 500):
    """Under Development..."""
    if reviewed:
        reviewed = 'true'
    else:
        reviewed = 'false'
    session = requests.Session()
    base_url = f'https://rest.uniprot.org/uniprotkb/search?&' \
               f'query=proteome:{identifier} AND (reviewed:{reviewed})&format=fasta&size={size}'
    response = session.get(base_url)
    total = int(response.headers['x-total-results'])
    fasta = [response.text]

    for _ in tqdm.tqdm(range(size, total, size)):
        try:
            response = session.get(response.links['next']['url'])
            fasta.extend([response.text])
        except KeyError:
            print(1)
            break

    fasta = '\n'.join(fasta)
    return fasta


def url_split(url_with_params):
    """Extract parameters from url 从网址中提取参数

    Args:
        url_with_params: URL with parameters 带参数的网址

    Returns:
        base_url: URL without parameters 没有参数的网址
        base_params: Separated parameters 被分离的参数
    """
    base_url = url_with_params[:re.search("\\?", url_with_params).start()]
    params = re.split("&", url_with_params[re.search("\\?", url_with_params).end():])
    params = [re.split("=", i) for i in params]
    base_params = dict()
    for parameter in params:
        base_params[parameter[0]] = parse.unquote(parameter[1])
    return base_url, base_params


def usi_split(usi):
    """Splitting USI"""
    component = dict()

    collection = re.search('(?<=mzspec:)\\S*?(?=:)', usi)

    component['collection'] = collection.group()

    split_point1 = collection.end() + 1
    split_point2 = re.search('(index|scan|trace|nativeId):\\d*', usi).end()

    component['msRun+index'] = usi[split_point1:split_point2]
    component['interpretation'] = usi[split_point2 + 1:]

    return component


def usi_detect(usi):
    """Detect if the string is USI"""
    try:
        c = usi_split(usi)
    except AttributeError:
        return False
    else:
        if re.search("pxd", c['collection'], re.I) and c['msRun+index'] and c['interpretation']:
            return True
        else:
            return False





def proteome_file_info_split(file_list: list):
    file_info = [[i,
                  re.match('UP\\d*', i).group(),
                  re.search('(?<=_)\\d{4}-\\d{2}-\\d{2}', i).group()
                  ] for i in file_list]
    file_info = [[i[0], i[1], time.strptime(i[2], '%Y-%m-%d')] for i in file_info]
    return file_info


def uniprot_proteome_id_detect(taxonomy_id):
    """Find proteome files based on UniProt taxonomy ID"""
    df = create_list_uniprot_proteome()
    target = df.loc[df['Tax_ID'] == str(taxonomy_id)]
    try:
        up_id = target['Proteome_ID'].values[0]
    except IndexError:
        raise IndexError('Taxonomy is not found.')
    return up_id


def proteome_file_detect(taxonomy_id):
    """Detection of the local presence of the corresponding proteomic fasta"""
    up_id = uniprot_proteome_id_detect(taxonomy_id)
    file_list = list_local_proteome()
    file_info = proteome_file_info_split(file_list)

    target_files = [i for i in file_info if i[1] == up_id]
    if not target_files:
        return False
    else:
        return True


def singleton(cls):
    """Singleton Pattern"""
    _instance = {}

    def inner():
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]

    return inner


def get_result_from_file(task=None):
    """Get Patpat search result from patpat_env/result/<task>/result.json

    Args:
        task: uuid

    Returns:
        dict, project metadata
    """
    if task:
        with open(f'patpat_env/result/{task}/result.json') as f:
            data_json = ''.join(f.readlines())
            data_dict = json.loads(data_json)
        f.close()

        return data_dict

    else:
        print('Need to input task uuid.')


def llj(data, step):
    """Grouping data"""
    if isinstance(data, list):
        if step < len(data):
            pass
        else:
            print(f'Warming: step too huge than data [{len(data)}=<{step}]')

        data_list = [data[i:i + step] for i in range(0, len(data), step)]
        return data_list
    else:
        raise ValueError('Input should be list.')



"""Logger module 日志记录器模块

The class CoreLogger is used to generate logs and search temporary files,
and is recommended to be used with hub.py and viewer.py.
class CoreLogger用于生成日志和搜索临时文件，建议配合hub.py和viewer.py使用。

    Typical usage example:
    典型用法示例：

    from patpat import logger

    l = logger.CoreLogger(uuid)
    l.set_core()
    l.set_tmp()
"""
import logging
import sys


class CoreLogger:
    """Core Log Module 核心日志模块

    Used to generate logs and search temporary files, recommended to use with hub.py and viewer.py.
    用于生成日志和搜索临时文件，
    """

    def __init__(self, uuid):
        """Initialization Class 初始化类

        Args:
            uuid: The uuid created by class MapperHub, or enter the uuid of each search task by yourself.
                类MapperHub创建的uuid，或自行输入曾经搜索任务的uuid。
        """
        self.id = uuid
        # Create the Logger object for log generation. 创建用于生成日志的Logger对象。
        self.logger = logging.getLogger('core')
        self.fmt = logging.Formatter('%(asctime)s\t%(levelname)s\t%(module)s\t%(funcName)s\t%(message)s')
        self.logger.setLevel(logging.DEBUG)

        # Create the Logger object used to generate the temporary file. 创建用于生成临时的Logger对象
        self.tmp = logging.getLogger('tmp')
        self.tmp.setLevel(logging.INFO)

    def set_core(self):
        """Used to set up two log processors, one sends logs to sys.stderr and the other sends them to the uuid.log file.
           用于设置两个日志处理器，一个将日志发送到sys.stderr，另一个将日志发送到uuid.log文件。
        """
        sh = logging.StreamHandler(stream=sys.stdout)
        sh.setFormatter(logging.Formatter('%(message)s'))
        sh.setLevel(logging.INFO)
        fh = logging.FileHandler(filename=f'patpat_env/logs/{self.id}.log', encoding='utf-8')
        fh.setFormatter(self.fmt)

        self.logger.addHandler(sh)
        self.logger.addHandler(fh)

    def set_tmp(self):
        """Processor used to set the log to be sent to the temporary file. 用于设置发送日志至临时文件的处理器。"""
        fh = logging.FileHandler(filename=f'patpat_env/tmp/{self.id}.log', encoding='utf-8')
        fh.setFormatter(logging.Formatter('%(message)s'))
        self.tmp.addHandler(fh)


class TestLogger:
    def __init__(self):
        self.logger = logging.getLogger('core')
        self.logger.setLevel(logging.DEBUG)

        self.sh = logging.StreamHandler(stream=sys.stdout)
        self.sh.setFormatter(logging.Formatter('%(message)s'))

        self.logger.addHandler(self.sh)



"""This module is used to check the connectivity of external proteomics databases 此模块用于确认外部蛋白质组学数据库的连通性

    Typical usage example:
    典型用法示例：

    from patpat import checker
    from patpat import retriever

    c = checker.GenericChecker()
    c.peptide_retrievers = [
        retriever.PridePeptideRetriever(),
        retriever.IProXPeptideRetriever()
    ]
    c.check()
"""
import requests
import tqdm

from . import retriever
from . import mapper


class GenericChecker:
    def check(self):
        raise NotImplementedError


class PrideChecker(GenericChecker):
    def __init__(self, times=3):
        self.retrievers = None
        self.times = times
        self.flag = {
            'project': 0,
            'protein': 0,
            'peptide': 0
        }
        self.message = {
            'project': [],
            'protein': [],
            'peptide': []
        }

    def _implement(self):
        retrievers = retriever.GenericPrideRetriever.__subclasses__()

        self.retrievers = [r() for r in retrievers]

    def _ans(self):
        print("\nPRIDE Connectivity Report")
        print(f"\tProject Retriever: {bool(self.flag['project'])}")
        print(f"\tProtein Retriever: {bool(self.flag['protein'])}")
        print(f"\tPeptide Retriever: {bool(self.flag['peptide'])}")

        if len([i for i in self.flag.values() if i]) == len(self.retrievers):
            print(f"Welcome to use PRIDEMapper")
            return mapper.PrideMapper()
        else:
            print(f"PRIDEMapper is not recommended")

    def check(self):
        self._implement()

        for r in self.retrievers:
            if isinstance(r, retriever.PrideProjectRetriever):
                flag, message = self._project_check(project_retriever=r)
                if flag:
                    self.flag['project'] = 1
                self.message['project'] = message

            elif isinstance(r, retriever.PrideProteinRetriever):
                flag, message = self._protein_check(protein_retriever=r)
                if flag:
                    self.flag['protein'] = 1
                self.message['protein'] = message

            elif isinstance(r, retriever.PridePeptideRetriever):
                flag, message = self._peptide_check(peptide_retriever=r)
                if flag:
                    self.flag['peptide'] = 1
                self.message['peptide'] = message
        m = self._ans()
        return m

    def _project_check(self, project_retriever):
        t = project_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the PRIDE Project API")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                c = requests.get(t.api, params=t.example)
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c.ok:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message

    def _protein_check(self, protein_retriever):
        t = protein_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the PRIDE Protein API")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                payloads = {'proteinAccession': t.example,
                            'pageSize': 5}
                c = requests.get(t.api, params=payloads)
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c.ok:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message

    def _peptide_check(self, peptide_retriever):
        t = peptide_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the PRIDE Peptide API")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                payloads = {'peptideSequence': t.example,
                            'pageSize': 5}
                c = requests.get(t.api, params=payloads)
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c.ok:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message


class IProXChecker(GenericChecker):
    def __init__(self, times=3):
        self.retrievers = None
        self.times = times
        self.flag = {
            'project': 0,
            'protein': 0,
            'peptide': 0
        }
        self.message = {
            'project': [],
            'protein': [],
            'peptide': []
        }

    def _implement(self):
        retrievers = retriever.GenericIProXRetriever.__subclasses__()

        self.retrievers = [r() for r in retrievers]

    def _ans(self):
        print("\niProX Connectivity Report")
        print(f"\tProject Retriever: {bool(self.flag['project'])}")
        print(f"\tProtein Retriever: {bool(self.flag['protein'])}")
        print(f"\tPeptide Retriever: {bool(self.flag['peptide'])}")

        if len([i for i in self.flag.values() if i]) == len(self.retrievers):
            print(f"Welcome to use IProXMapper")
            return mapper.IProXMapper()
        else:
            print(f"IProXMapper is not recommended")

    def check(self):
        self._implement()

        for r in self.retrievers:
            if isinstance(r, retriever.IProXProjectRetriever):
                flag, message = self._project_check(project_retriever=r)
                if flag:
                    self.flag['project'] = 1
                self.message['project'] = message

            elif isinstance(r, retriever.IProXProteinRetriever):
                flag, message = self._protein_check(protein_retriever=r)
                if flag:
                    self.flag['protein'] = 1
                self.message['protein'] = message

            elif isinstance(r, retriever.IProXPeptideRetriever):
                flag, message = self._peptide_check(peptide_retriever=r)
                if flag:
                    self.flag['peptide'] = 1
                self.message['peptide'] = message
        m = self._ans()
        return m

    def _project_check(self, project_retriever):
        t = project_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the iProX Project API")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                c = requests.get("/".join([t.api, t.example]))
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c.ok:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message

    def _protein_check(self, protein_retriever):
        t = protein_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the iProX Protein API")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                payloads = {'proteinAccession': t.example,
                            'pageNumber': '1',
                            'pageSize': '5',
                            'resultType': 'compact'}
                c = requests.get(t.api, params=payloads)
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c.ok:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message

    def _peptide_check(self, peptide_retriever):
        t = peptide_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the iProX Peptide API")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                payloads = {'peptideSequence': t.example,
                            'pageNumber': '1',
                            'pageSize': '200',
                            'resultType': 'compact'}
                c = requests.get(t.api, params=payloads)
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c.ok:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message


class MassIVEChecker(GenericChecker):
    def __init__(self, times=3):
        self.retrievers = None
        self.times = times
        self.flag = {
            'project': 0,
            'protein': 0,
            'peptide': 0,
            'protein2project': 0,
            'peptide2project': 0,
        }
        self.message = {
            'project': [],
            'protein': [],
            'peptide': [],
            'protein2project': [],
            'peptide2project': [],
        }

    def _implement(self):
        retrievers = retriever.GenericMassIVERetriever.__subclasses__()

        self.retrievers = [r() for r in retrievers]

    def _ans(self):
        print("\nMassIVE Connectivity Report")
        print(f"\tProject Retriever: {bool(self.flag['project'])}")
        print(f"\tProtein Retriever: {bool(self.flag['protein'])}")
        print(f"\tPeptide Retriever: {bool(self.flag['peptide'])}")
        print(f"\tprotein2project Retriever: {bool(self.flag['protein2project'])}")
        print(f"\tpeptide2project Retriever: {bool(self.flag['peptide2project'])}")

        if len([i for i in self.flag.values() if i]) == len(self.retrievers):
            print(f"Welcome to use MassIVEMapper")
            return mapper.MassIVEMapper()
        else:
            print(f"MassIVEMapper is not recommended")

    def check(self):
        self._implement()

        for r in self.retrievers:
            if isinstance(r, retriever.MassIVEProjectRetriever):
                flag, message = self._check(this_retriever=r)
                if flag:
                    self.flag['project'] = 1
                self.message['project'] = message

            elif isinstance(r, retriever.MassIVEProteinRetriever):
                flag, message = self._check(this_retriever=r)
                if flag:
                    self.flag['protein'] = 1
                self.message['protein'] = message

            elif isinstance(r, retriever.MassIVEPeptideRetriever):
                flag, message = self._check(this_retriever=r)
                if flag:
                    self.flag['peptide'] = 1
                self.message['peptide'] = message

            elif isinstance(r, retriever.MassIVEPro2ProjectRetriever):
                flag, message = self._check(this_retriever=r)
                if flag:
                    self.flag['protein2project'] = 1
                self.message['protein2project'] = message

            elif isinstance(r, retriever.MassIVEPep2ProjectRetriever):
                flag, message = self._check(this_retriever=r)
                if flag:
                    self.flag['peptide2project'] = 1
                self.message['peptide2project'] = message
        m = self._ans()
        return m

    def _check(self, this_retriever):
        t = this_retriever
        flag = False
        message = []

        print(f"\nCheck the connectivity of the {t.__class__.__name__}")
        for _ in tqdm.tqdm(range(self.times)):
            try:
                t.request_word = t.example
                c = t.url_requester()
            except requests.exceptions.RequestException as e:
                message.extend([e])
                continue
            else:
                if c:
                    flag = True
                    break
                else:
                    message.extend([c.text])
        if flag:
            return flag, message
        else:
            return flag, message



"""This module is used to build the configuration information that needs to be retrieved.
   该模块用于建立需要检索的配置信息。

The configuration information required to complete the mapping are the protein metadata and the peptide sequences
to be retrieved. The necessary protein metadata are: the protein identifier, the species or tissue to which
the protein belongs, and the protein sequence, where the protein sequence is used to generate the peptide sequences
to be retrieved.
完成映射所需要的配置信息有蛋白质的元数据和需要检索的肽段序列。
必要的蛋白质元数据有：蛋白质识别符、蛋白质所属物种或组织、蛋白质序列，其中蛋白质序列用于产生需要检索的肽段序列。

    Typical usage example:
    典型用法示例：

    from patpat import querier

    p1 = querier.UniProtProteinQuerier()
    p1.identifier = 'E9PV96'
    p1.query()
    _, organism_, fasta_ = p1.get_properties()

    p2 = querier.LocalPeptideQuerier()
    p2.set_params(sequence=fasta_['sequence'],
                  organism=organism_)
    p2.query()
    digestion_params_, source_, filtered_peptides_ = p2.get_properties()

"""

import collections
import re

import requests
import tqdm
from pyteomics import parser, fasta

from . import utility


class ProteinQuerier(object):
    """Base class for implementing the protein metadata search function. 实现蛋白质元数据搜索功能的基类。

    This base class gives preliminary interface definitions to build subclasses against protein databases.
    这个基类给出了初步的接口定义，以便针对蛋白质数据库建立子类。
    """

    def __init__(self):
        # Structuring protein fasta into a dictionary. 将蛋白质fasta结构化为字典。
        self.fasta: dict = {
            'description': '',
            'sequence': ''
        }
        self.organism: dict = dict()  # the species or tissue to which the protein belongs 蛋白质物种或组织信息
        self.identifier: str = ''  # Identifier of protein 蛋白质的识别符
        self._headers: dict = {'Accept': 'application/json'}

    def _request_api(self, *args):
        """Functions that actually interact with the protein database. 实际与蛋白质数据库交互的函数。"""
        raise NotImplementedError

    def set_params(self, *args):
        """Functions for setting query parameters. 设置查询参数的函数。"""
        raise NotImplementedError

    def get_properties(self):
        """Functions to get query results. 获取查询结果的函数。"""
        raise NotImplementedError

    def query(self):
        """Functions for external calls. 供外部调用的函数。"""
        raise NotImplementedError


class PeptideQuerier(object):
    """Generate the base class of peptide sequences to be retrieved. 生成要检索的肽段序列的基类。

    This base class gives preliminary interface definitions to create subclasses for the peptide generation methods.
    这个基类给出了初步的接口定义，以便针对肽段生成方法建立子类。
    """

    def __init__(self):
        self.sequence: str = ''  # protein sequence 蛋白质序列
        self.organism: dict = dict()  # the species or tissue to which the protein belongs 蛋白质物种或组织信息

        # Filtered peptides for output to external 经过筛选的肽段，可输出至外部
        self.filtered_peptides: list = []

        # Unfiltered full peptides generated from protein sequences 根据蛋白质序列产生的未经过滤全部肽段
        self._peptides: list = []

    def set_params(self, *args):
        """Functions for setting query parameters. 设置查询参数的函数。"""
        raise NotImplementedError

    def get_properties(self):
        """Functions to get query results. 获取查询结果的函数。"""
        raise NotImplementedError

    def query(self):
        """Functions for external calls. 供外部调用的函数。"""
        raise NotImplementedError


class UniProtProteinQuerier(ProteinQuerier):
    """ProteinQuerier subclass for querying protein metadata through the UniProt database.
       通过UniProt数据库查询蛋白质元数据的ProteinQuerier子类。
    """

    def __init__(self):
        super().__init__()

    def _request_api(self, accession):
        """Functions that actually interact with the UniProt database. 实际与UniProt数据库交互的函数。

        Args:
            accession: str. UniProt accession of protein 蛋白质的UniProt编号

        Returns:
            protein_fasta: str. fasta of protein 蛋白质fasta

        """
        acc = accession
        url = f"https://rest.uniprot.org/uniprotkb/stream?compressed=false&format=fasta&query={acc}"
        protein_fasta = requests.get(url,
                                     headers=self._headers).text
        return protein_fasta

    def set_params(self, accession):
        """Functions for setting query parameters. 设置查询参数的函数。

        Args:
            accession: str. UniProt accession of protein 蛋白质的UniProt编号

        Returns:
            None

        Updates:
            self.identifier
        """
        if re.match('[OPQ]\\d[A-Z\\d]{3}\\d|[A-NR-Z]\\d([A-Z][A-Z\\d]{2}\\d){1,2}', accession).group() == accession:
            self.identifier = accession
        else:
            raise TypeError('Can not identify identifier, please input UniProt accession.')

    def query(self):
        """Functions for external calls. 供外部调用的函数。

        Returns:
            None

        Updates:
            self.fasta
            self.organism

        Raises:
            ValueError: self.identifier is empty.
        """
        if not self.identifier:
            raise ValueError('Please run func self.set_params() first. 请先运行self.set_params()函数')

        protein_fasta = self._request_api(self.identifier)
        fasta_sep = re.split('\n', protein_fasta)
        self.fasta['description'] = fasta_sep[0]
        self.fasta['sequence'] = ''.join(fasta_sep[1:])

        self.organism['name'] = re.search('(?<=OS=).*(?=OX=)',
                                          self.fasta['description']).group()[:-1]
        self.organism['accession'] = re.search('(?<=OX=).*(?=GN=)',
                                               self.fasta['description']).group()[:-1]

    def get_properties(self):
        """Functions to get query results. 获取查询结果的函数。

        Returns:
            self.identifier: str. UniProt accession of protein 蛋白质的UniProt编号
            self.organism: dict. UniProt taxonomy of protein 蛋白质所属的UniProt物种分类
            self.fasta: dict.

        Raises:
            ValueError: self.identifier is empty
        """
        if self.identifier:
            return self.identifier, self.organism, self.fasta
        elif self.identifier is None:
            raise ValueError('Please run func self.set_params() first. 请先运行self.set_params()函数')


class LocalPeptideQuerier(PeptideQuerier):
    """The PeptideQuerier subclass used to generate the peptide sequences by local library building.
       通过本地建库的方法产生肽段序列的PeptideQuerier子类。
    """

    def __init__(self):
        super().__init__()

        self.digestion_params = None  # Parameters of protein in-silico enzymatic 蛋白质模拟酶切的参数
        self.source = None  # Local proteome file address 本地蛋白质组文件地址
        self._searched_peptides = None  # Peptide information after local database search 经过本地数据库搜索的肽段信息
        self._base = None  # Local Peptide Database 本地肽段数据库

    def set_params(self,
                   sequence,
                   organism,
                   digestion_params: dict = None,
                   source: str = None):
        """Functions for setting query parameters. 设置查询参数的函数。

        Args:
            sequence: str. protein sequence 蛋白质序列
            organism: dict. the species or tissue to which the protein belongs 蛋白质物种或组织信息
            digestion_params: dict. Parameters of protein in-silico enzymatic 蛋白质模拟酶切的参数
            source: str. Local proteome file address 本地蛋白质组文件地址

        Returns:
            None

        Updates:
            self.sequence
            self.organism
            self.digestion_params
            self.source
        """
        self.sequence = sequence
        self.organism = organism

        if digestion_params is None:
            digestion_params = {
                'rules': 'trypsin',
                'miss': 1,
                'min_length': 7,
                'max_length': 35}

        self.digestion_params = digestion_params
        self.source = source

    def digestion(self):
        """Functions for protein in-silico enzymatic 用于蛋白质模拟酶切的函数

        Returns:
            None

        Updates:
            self._peptides
        """
        self._peptides = list(parser.cleave(self.sequence,
                                            rule=self.digestion_params['rules'],
                                            missed_cleavages=self.digestion_params['miss'],
                                            min_length=self.digestion_params['min_length'],
                                            max_length=self.digestion_params['max_length']))

    def search(self, threshold=1):
        """Functions for local library search. 用于本地数据库搜索的函数

        Args:
            threshold: int. Filtering threshold, which refers to the peptide is shared by how many proteins,
                            greater than the threshold is filtered
                            过滤阈值，指被多少个蛋白质共享的肽段，大于阈值则被过滤

        Returns:
            None

        Updates:
            self._searched_peptides
            self._base
            self.filtered_peptides
        """
        print('Choose local peptide search.')

        source = self._select_local_proteome()

        print(f'Proteome file: {source}')

        f = fasta.read(source)
        input_peps = self._peptides
        try:
            rule = self.digestion_params['rules']
            miss = self.digestion_params['miss']
            min_length = self.digestion_params['min_length']
            max_length = self.digestion_params['max_length']

        except ValueError:
            raise ValueError('digestion() params error!')

        base = collections.defaultdict(list)
        print('Creating database...')
        for protein in tqdm.tqdm(f):
            peptides = parser.cleave(protein.sequence,
                                     rule=rule,
                                     missed_cleavages=miss,
                                     min_length=min_length,
                                     max_length=max_length)
            for peptide in peptides:
                try:
                    base[peptide][0] += 1
                    base[peptide][1].add(protein.description)
                except IndexError:
                    base[peptide] = [1, {protein.description}]

        ans = collections.defaultdict(list)
        for pep in input_peps:
            try:
                ans[pep].append(base[pep][0])
                ans[pep].append(base[pep][1])
            except IndexError:
                ans[pep] = [0, 'loss']
        ans = sorted(zip(ans.values(), ans.keys()))
        ans = [[j for j in utility.flatten(i)] for i in ans]
        errs = [i[2] for i in ans if i[0] == 0]
        if len(errs):
            print(f'There are some peptides that are not in the query database: {errs}')

        self._searched_peptides = ans
        self._base = base
        self._search_filter(threshold=threshold)

    def _select_local_proteome(self):
        """The function is used to automatically search for locally available proteome file
           此函数用于自动化搜索本地可用的蛋白质组文件

        Returns:
            None

        Updates:
            self.source
        """
        source = self.source

        if source is None:
            taxonomy_id = self.organism['accession']
            up_id = utility.uniprot_proteome_id_detect(taxonomy_id)
            file_list = utility.list_local_proteome()
            file_info = utility.proteome_file_info_split(file_list)

            target_files = [i for i in file_info if i[1] == up_id]
            if not target_files:
                print(f"The {self.organism['name']} {up_id} proteome file was not found locally.")

                flag = input('Do you want to download it?(y/n)')
                if flag == 'y':
                    source = utility.download_uniprot_opg(taxonomy_id)
                else:
                    raise FileNotFoundError(f'User cancel operation.')

            else:
                target_file = [i[0] for i in target_files if i[2] == max([j[2] for j in target_files])][0]
                source = f'patpat_env/proteome/uniprot-proteome_{target_file}.fasta'

        else:
            source = f'patpat_env/proteome/{source}'

        self.source = source

        return source

    def _search_filter(self, threshold=1):
        """The function is used to filter peptide sequences that are larger than the threshold
           本函数用于过滤大于阈值的肽段序列
        """
        ans = self._searched_peptides

        ans = [i[2] for i in ans if i[0] <= threshold and i[0] != 0]
        self.filtered_peptides = ans

    def get_properties(self):
        """Functions to get query results. 获取查询结果的函数。

        Returns:
            self.digestion_params: dict. Parameters of protein in-silico enzymatic 蛋白质模拟酶切的参数
            self.source: str. Local proteome file address 本地蛋白质组文件地址
            self.filtered_peptides: list. Filtered peptides for output to external 经过筛选的肽段，可输出至外部

        Raises:
            ValueError: self.sequence or self.organism is empty.
        """
        if not self.sequence or not self.organism:
            raise ValueError('Please run func self.set_params() first. 请先运行self.set_params()函数')

        return self.digestion_params, self.source, self.filtered_peptides

    def query(self, threshold=1):
        """Functions for external calls. 供外部调用的函数。

        Raises:
            ValueError: self.sequence or self.organism is empty.
        """
        if not self.sequence or not self.organism:
            raise ValueError('Please run func self.set_params() first. 请先运行self.set_params()函数')

        self.digestion()
        self.search(threshold=1)


class UniProtPeptideQuerier(PeptideQuerier):
    """Under Development..."""

    def __init__(self,
                 sequence,
                 digestion_params: dict = None,
                 search_args=None):
        super().__init__()
        self.sequence = sequence

        if digestion_params is None:
            digestion_params = {
                'rules': 'trypsin',
                'miss': 1,
                'min_length': 7,
                'max_length': 35}

        self.digestion_params = digestion_params
        self.uniprot_job = dict()
        self.search_args = search_args
        self.source = 'UNIPROT'

        self._searched_peptides = None
        self._digestion_params = None
        self._peptides = None

        self.filtered_peptides = None

        # CALL ERROR
        raise NotImplementedError("This class is under development...")

    def set_params(self):
        pass

    def digestion(self):
        self._peptides = list(parser.cleave(self.sequence,
                                            rule=self.digestion_params['rules'],
                                            missed_cleavages=self.digestion_params['miss'],
                                            min_length=self.digestion_params['min_length'],
                                            max_length=self.digestion_params['max_length']))

    def search(self):
        print('Choose UniProt peptide search.')

    def _uniprot_peptide_search_request(self):
        print('job id={}'.format(self.uniprot_job['job']))

    def _uniprot_peptide_search_check(self):
        # print("This took: {} sec".format(int(end - start)))
        pass

    def _uniprot_peptide_search_answer(self):
        pass

    def _uniprot_search_filter(self, *args, threshold=1):
        pass

    def get_properties(self):
        return self.digestion_params, self.source, self.filtered_peptides

    def query(self):
        self.digestion()
        self.search()


if __name__ == '__main__':
    identifier_ = 'E9PV96'
    p1 = UniProtProteinQuerier()
    p1.set_params(accession=identifier_)
    p1.query()
    _, organism_, fasta_ = p1.get_properties()

    p2 = LocalPeptideQuerier()
    p2.set_params(sequence=fasta_['sequence'],
                  organism=organism_)
    p2.query()
    digestion_params_, source_, filtered_peptides_ = p2.get_properties()



"""This module contains classes for interacting with external proteomics databases.此模块包含与外部蛋白质组学数据库交互的类。

The classes in this module use three levels of inheritance. First,(1) all subclasses need to inherit from
the Retriever base class, second,(2) depending on the database to be targeted, a base class  needs to be
created for that database, and finally, the subclass that collects the specified data from that database needs
to inherit from that database's base class (3). For example, the base class of PRIDE database is named
GenericPrideRetriever (2), which inherits Retriever (1) base class, and the subclass that collects protein
information from PRIDE is named PrideProteinRetriever (3).
本模块的类使用了三层继承。首先，所有的子类都需要继承Retriever基类（1），其次，根据要针对的数据库，需要为该数据库创建一个基类（2），
最后，从该数据库收集指定数据的子类（3）需要继承该数据库的基类。例如，PRIDE数据库的基类名为GenericPrideRetriever（2），
它继承了Retriever基类（1），从PRIDE收集蛋白质信息的子类名为PrideProteinRetriever（3）

    Typical usage example:
    典型用法示例：

    from patpat import retriever

    r = retriever.PrideProteinRetriever()
    r.request_word = "Q9CWY9"
    r.get_payloads_on_web()
    r.payloads["pageSize"] = 200 # Optional: Parameters can be modified 可选：参数可修改
    r.retrieve()
"""

import re
import time
import logging
import random

import requests

from . import utility


class Retriever(object):
    """Base class for accessing proteomics database URL. 用于访问蛋白质组数据库网址的基类。

    Common proteomics databases are PRIDE, PeptideAtlas, MassIVE, jPOST, iProx, Panorama Public.
    This base class gives preliminary interface definitions to build base classes against databases.
    常见的蛋白质组学数据库有 PRIDE、PeptideAtlas、MassIVE、jPOST、iProx、Panorama Public。
    这个基类给出了初步的接口定义，以便针对数据库建立基类。

    """

    def retrieve(self):
        """The call makes the class Retriever  run and structures the incoming data.
           该调用使Retriever类运行，并对传入的数据进行结构化。

        This is the most important method of the class, so please design it according to your actual needs!
        It is worth noting that this method is required whether the class is called directly,
        or subsequently through another class.
        这是该类中最重要的方法，所以请根据你的实际需要来设计它。值得注意的是，不管是直接调用该类，还是随后通过其他类调用，都需要这个方法。
        """
        raise NotImplementedError

    def url_requester(self):
        """For direct access to the database URL. 用于直接访问数据库URL。

        Access parameters can be set according to actual needs.
        可以根据实际需要设置访问参数。
        """
        raise NotImplementedError

    def get_payloads_on_web(self):
        """Used to get the list of optional parameters from the server of the database.用于从数据库的服务器中获取可选参数列表。

        Optionally, according to the actual needs of the database using.
        可选的，根据数据库实际需要采用。
        """
        raise NotImplementedError


class GenericPrideRetriever(Retriever):
    """Generic base class for PRIDE database. 针对PRIDE数据库的通用基类。

    """

    def __init__(self):
        self._request_word = ''  # Keywords to be requested 需要请求的关键词
        self.headers = None  # HTTP protocol headers HTTP协议的标头
        self.payloads = None  # parameters. 参数列表
        self.url = ''  # URL 网址

    def retrieve(self):
        """Call the methods run by the class PrideRetriever. 调用PrideRetriever类运行的方法。"""
        raise NotImplementedError

    def url_requester(self):
        """Direct access to PRIDE database API. 直接访问PRIDE数据库API的方法。"""
        raise NotImplementedError

    def get_payloads_on_web(self):
        """Access PRIDE API and get the list of parameters of the URL through redirection.
           访问PRIDE API并通过重定向获得URL的参数列表。
        """
        raise NotImplementedError

    def check_end_page(self, url, payloads):
        """Determine if the page is the end page. 确定该页是否为结束页。

        Determine if the page is an end page by the presence of the returned request.json()['_embedded'].
        通过回传的request.json()['_embedded']是否存在，判断该页是否为尾页。

        Args:
            url: URL without parameters 没有参数的网址
            payloads: Parameters to be entered 需要输入的参数

        Returns:
            Bool

        Raises:
            KeyError: Catching this error means that request.json()['_embedded'] does not exist,
             i.e. the page is the end page. 捕捉到这个错误意味着request.json()['_embedded']不存在，也就是说，该页面是结束页。
        """
        page = payloads['page']

        # Almost all of the if-else judgments in this function are used to determine which page number should be
        # written to the log.
        # 这个函数中几乎所有的if-else判断都是用来确定哪一个页码应该被写入日志的。
        if int(page) == 0:
            logging.getLogger('core').debug("----------------------------------------")
            logging.getLogger('core').debug("\tTest if the {} page is the end page.".format(str(int(page))))
        else:
            logging.getLogger('core').debug("----------------------------------------")
            logging.getLogger('core').debug("\tTest if the {} page is the end page.".format(str(int(page) - 1)))

        n = 0
        end = 3
        for n in range(0, end, 1):
            d = requests.get(url, params=payloads)
            time.sleep(random.random())  # Take a break and avoid frequent access. 歇歇，避免频繁访问。
            if d.ok:
                try:
                    _ = d.json()['_embedded']
                except KeyError:
                    pass
                else:
                    logging.getLogger('core').debug("\tthe {} page is not the end page.".format(str(int(page) - 1)))
                    logging.getLogger('core').debug("----------------------------------------")
                    return False
        if n == end - 1 and int(page) != 0:
            logging.getLogger('core').debug("\tthe {} page is the end page.".format(str(int(page) - 1)))
            logging.getLogger('core').debug("----------------------------------------")
            return True
        elif n == end - 1 and int(page) == 0:
            logging.getLogger('core').debug("\tthe {} page is the end page.".format(str(int(page))))
            logging.getLogger('core').debug("----------------------------------------")
            return True


class PrideProjectRetriever(GenericPrideRetriever):
    """Used to collect project information in PRIDE database. 用于收集PRIDE数据库中的项目信息。"""

    def __init__(self):
        super().__init__()
        self.response = dict()  # Collect the returned data through this property. 通过这个属性收集返回的数据。
        self.api = 'https://www.ebi.ac.uk/pride/ws/archive/v2/projects'
        self.example = 'PXD003415'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Only allows searching for items assigned a number by ProteomeXchange.
        仅允许搜索被ProteomeXchange分配编号的项目。

        Args:
            request_word: keyword 请求关键词
        """
        if re.search("pxd", request_word, re.I) is None:
            raise TypeError('Input error! 输入错误！')
        else:
            self._request_word = request_word

    def retrieve(self):
        """Used to request project data from PRIDE database. 用于向PRIDE数据库请求项目数据。

        Returns:
            None

        Updates:
             self.response

        """
        url_response = self.url_requester()
        logging.getLogger('core').debug('query:{}'.format(self._request_word))

        if url_response.ok:
            url_response = url_response.json()
        else:
            logging.error('No correct return transmission received! 未收到正确回传！')
            url_response = dict()

        time.sleep(.3)
        self.response = url_response
        return self.response

    def url_requester(self):
        """API for direct access to PRIDE database project data. 用于直接访问PRIDE数据库项目数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}

        self.url = "/".join((self.api, self._request_word))

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads,
                                    timeout=120  # Avoid blocking
                                    )

        return url_response

    def get_payloads_on_web(self):
        """Occupy a seat 占位"""
        raise NotImplementedError("Don\'t need that in this class. 该类不需要这个！")


class PridePeptideRetriever(GenericPrideRetriever):
    """Used to collect peptide information in PRIDE database. 用于收集PRIDE数据库中的肽段信息。"""

    def __init__(self):
        super().__init__()
        self.response = dict()  # Collect the returned data through this property. 通过这个属性收集返回的数据。
        self.api = "https://www.ebi.ac.uk/pride/ws/archive/v2/spectra"
        self.example = 'TCVADESAENCDK'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word

        """
        self._request_word = request_word

    def retrieve(self):
        """Used to request peptide data from PRIDE database. 用于向PRIDE数据库请求肽段数据。

        Returns:
            Updated self.response

        """
        url = self.url
        payloads = self.payloads
        headers = self.headers
        response = dict()
        flag = False

        logging.getLogger('core').info('query:{}'.format(self._request_word))
        while flag is False:
            page = payloads["page"]
            url_response = requests.get(url, headers=headers, params=payloads,
                                        timeout=60  # Avoid blocking
                                        ).json()
            try:
                response["page:{}".format(page)] = {'spectraevidences': url_response['_embedded']['spectraevidences']}
            except KeyError:
                flag = True
                # flag = self.check_end_page(url, payloads)
            else:
                response["page:{}".format(page)]['link'] = url_response["_links"]["self"]["href"]
                logging.getLogger('core').info("page {} is completed.".format(page))
            finally:
                payloads["page"] = str(int(page) + 1)
        self.response = response
        return self.response

    def url_requester(self):
        """API for direct access to PRIDE database peptide data. 用于直接访问PRIDE数据库肽段数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {"peptideSequence": self._request_word,
                             "pageSize": 200}

        self.url = self.api

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads)
        return url_response

    def get_payloads_on_web(self):
        """Access the URL and get the list of parameters of the URL through redirection.
           访问URL并通过重定向获得URL的参数列表。
        """
        url_response = self.url_requester()

        if url_response.ok:
            url_response = url_response.json()
            self.url, self.payloads = utility.url_split(url_response["_links"]['self']['href'])

        return url_response


class PrideProteinRetriever(GenericPrideRetriever):
    """Used to collect protein information in PRIDE database. 用于收集PRIDE数据库中的蛋白质信息。"""

    def __init__(self):
        super().__init__()
        self.response = dict()
        self.api = "https://www.ebi.ac.uk/pride/ws/archive/v2/peptideevidences"
        self.example = 'E9PV96'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Only UniProt numbers are allowed.
        仅允许UniProt编号

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word

        """
        if re.match('[OPQ]\\d[A-Z\\d]{3}\\d|[A-NR-Z]\\d([A-Z][A-Z\\d]{2}\\d){1,2}',
                    request_word).group() == request_word:
            self._request_word = request_word
        else:
            raise TypeError('Input error! 输入错误！')

    def retrieve(self):
        """Used to request peptide data from PRIDE database. 用于向PRIDE数据库请求蛋白质数据。

        Returns:
            None

        Updates:
            self.response
        """
        url = self.url
        payloads = self.payloads
        headers = self.headers
        response = dict()
        flag = False

        logging.getLogger('core').info('query:{}'.format(self._request_word))
        while flag is False:
            try:
                page = payloads['page']
            except KeyError as e:
                logging.getLogger('core').error(e)
                logging.getLogger('core').error('Reply 1 time')
                self.get_payloads_on_web()
                page = payloads['page']

            url_response = requests.get(url, headers=headers, params=payloads,
                                        timeout=120  # Avoid blocking
                                        ).json()
            try:
                response['page:{}'.format(page)] = {'peptideevidences': url_response['_embedded']['peptideevidences']}
            except KeyError:
                flag = True
                # flag = self.check_end_page(url, payloads)
            else:
                response['page:{}'.format(page)]['link'] = url_response["_links"]["self"]["href"]
                logging.getLogger('core').info('page {} is completed.'.format(page))
            finally:
                payloads['page'] = str(int(page) + 1)
        self.response = response

    def url_requester(self):
        """API for direct access to PRIDE database protein data. 用于直接访问PRIDE数据库蛋白质数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {'proteinAccession': self._request_word,
                             'pageSize': 200}

        self.url = self.api

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads)
        return url_response

    def get_payloads_on_web(self):
        """Access the URL and get the list of parameters of the URL through redirection
           访问URL并通过重定向获得URL的参数列表
        """
        url_response = self.url_requester()

        if url_response.ok:
            url_response = url_response.json()
            self.url, self.payloads = utility.url_split(url_response["_links"]['self']['href'])


class GenericIProXRetriever(Retriever):
    """Generic base class for iProX database. 针对iProX数据库的通用基类。"""

    def __init__(self):
        self._request_word = ''  # Keywords to be requested 需要请求的关键词
        self.headers = None  # HTTP protocol headers HTTP协议的标头
        self.payloads = None  # parameters. 参数列表
        self.url = ''  # URL 网址

    def retrieve(self):
        """Call the methods run by the class IProXRetriever. 调用IProXRetriever类运行的方法。"""
        raise NotImplementedError

    def url_requester(self):
        """Direct access to iProX database API. 直接访问iProX数据库API的方法。"""
        raise NotImplementedError

    def get_payloads_on_web(self):
        """Access iProX API and get the list of parameters of the URL through redirection.
           访问iProX API并通过重定向获得URL的参数列表。
        """
        raise NotImplementedError


class IProXProjectRetriever(GenericIProXRetriever):
    """Used to collect project information in iProX database. 用于收集iProX数据库中的项目信息。"""

    def __init__(self):
        super().__init__()
        self.response = dict()
        self.api = 'https://www.iprox.cn/proxi/datasets'
        self.example = 'PXD006512'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Only allows searching for items assigned a number by ProteomeXchange.
        仅允许搜索被ProteomeXchange分配编号的项目。

        Args:
            request_word: keyword 请求关键词
        """
        if re.search("pxd", request_word, re.I) is None:
            raise TypeError("输入错误")
        else:
            self._request_word = request_word

    def retrieve(self):
        """Used to request project data from iProX database. 用于向iProX数据库请求项目数据。

        Returns:
            None

        Updates:
            self.response
        """
        url_response = self.url_requester()
        logging.getLogger('core').debug('query:{}'.format(self._request_word))

        if url_response.ok:
            url_response = url_response.json()
        else:
            logging.error('No correct return transmission received! 未收到正确回传！')
            url_response = dict()
        time.sleep(.3)
        self.response = url_response

    def url_requester(self):
        """API for direct access to iProX database project data. 用于直接访问iProX数据库项目数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}

        self.url = "/".join((self.api, self._request_word))

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads,
                                    timeout=120  # Avoid blocking
                                    )
        return url_response

    def get_payloads_on_web(self):
        """Occupy a seat 占位"""
        raise NotImplementedError("Don\'t need that in this class. 该类不需要这个！")


class IProXPeptideRetriever(GenericIProXRetriever):
    """Used to collect peptide information in iProX database. 用于收集iProX数据库中的肽段信息。"""

    def __init__(self):
        super().__init__()
        self.response = dict()
        self.api = "https://www.iprox.cn/proxi/spectra"
        self.example = 'TCVADESAENCDK'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word
        """
        if len(request_word) < 7:
            raise TypeError("输入错误")
        else:
            self._request_word = request_word

    def retrieve(self):
        """Used to request peptide data from iProX database. 用于向iProX数据库请求肽段数据。

        Returns:
            None

        Updates:
            self.response
        """
        url = self.url
        payloads = self.payloads
        headers = self.headers
        response = dict()
        flag = False

        logging.getLogger('core').info(f'query:{self._request_word}')
        while flag is False:
            page = payloads['pageNumber']
            url_response = requests.get(url, headers=headers, params=payloads,
                                        timeout=60  # Avoid blocking
                                        )

            if url_response.ok:
                if url_response.json():
                    response[f'page:{page}'] = url_response.json()
                    logging.getLogger('core').info("page {} is completed.".format(page))
                else:
                    flag = True
            else:
                logging.getLogger('core').error(url_response.json())
                raise requests.exceptions.RequestException(url_response.json())
            payloads['pageNumber'] = str(int(page) + 1)

        self.response = response

    def url_requester(self):
        """API for direct access to iProX database peptide data. 用于直接访问iProX数据库肽段数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {"Accept": "application/json"}
        if self.payloads is None:
            self.payloads = {'peptideSequence': self._request_word,
                             'pageNumber': '1',
                             'pageSize': '200',
                             'resultType': 'compact'}

        self.url = self.api

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads)
        return url_response

    def get_payloads_on_web(self):
        """Access the URL and get the list of parameters of the URL through redirection.
           访问URL并通过重定向获得URL的参数列表。
        """
        url_response = self.url_requester()

        if url_response.ok:
            url_response = url_response
            self.url, self.payloads = utility.url_split(url_response.url)

        return url_response


class IProXProteinRetriever(GenericIProXRetriever):
    """Used to collect protein information in iProX database. 用于收集iProX数据库中的蛋白质信息。"""

    def __init__(self):
        super().__init__()
        self.response = dict()
        self.api = 'https://www.iprox.cn/proxi/psms'
        self.example = 'P60709'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Only UniProt numbers are allowed.
        仅允许UniProt编号

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word
        """
        if re.match('[OPQ]\\d[A-Z\\d]{3}\\d|[A-NR-Z]\\d([A-Z][A-Z\\d]{2}\\d){1,2}',
                    request_word).group() == request_word:
            self._request_word = request_word
        else:
            raise TypeError('Input error! 输入错误！')

    def retrieve(self):
        """Used to request protein data from iProX database. 用于向iProX数据库请求蛋白质数据。

        Returns:
            Updated self.response

        """
        url = self.url
        payloads = self.payloads
        headers = self.headers
        response = dict()
        flag = False

        logging.getLogger('core').info(f'query:{self._request_word}')

        while flag is False:
            page = payloads['pageNumber']
            url_response = requests.get(url, headers=headers, params=payloads,
                                        timeout=120  # Avoid blocking
                                        )

            if url_response.ok:
                if url_response.json():
                    response[f'page:{page}'] = url_response.json()
                    logging.getLogger('core').info("page {} is completed.".format(page))
                else:
                    flag = True
            else:
                logging.getLogger('core').error(url_response.json())
                raise requests.exceptions.RequestException(url_response.json())
            payloads['pageNumber'] = str(int(page) + 1)

        self.response = response

    def url_requester(self):
        """API for direct access to iProX database protein data. 用于直接访问iProX数据库蛋白质数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {'proteinAccession': self._request_word,
                             'pageNumber': '1',
                             'pageSize': '200',
                             'resultType': 'compact'}

        self.url = self.api

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads)
        return url_response

    def get_payloads_on_web(self):
        """Access the URL and get the list of parameters of the URL through redirection
           访问URL并通过重定向获得URL的参数列表
        """
        url_response = self.url_requester()

        if url_response.ok:
            url_response = url_response
            self.url, self.payloads = utility.url_split(url_response.url)


class GenericMassIVERetriever(Retriever):
    """Generic base class for MassIVE database. 针对 MassIVE 数据库的通用基类。"""

    def __init__(self):
        self._request_word = ''  # Keywords to be requested 需要请求的关键词
        self.headers = None      # HTTP protocol headers HTTP协议的标头
        self.payloads = None     # parameters. 参数列表
        self.url = ''            # URL 网址
        self.response = None     # Collect the returned data through this property. 通过这个属性收集返回的数据。

        self.api = ''
        self._total = None       # total rows in this retrieve
        self._current = None     # current page
        self._offsets = None     # threshold of page

    def retrieve(self):
        """Used to request data from MassIVE database. 用于向 MassIVE 数据库请求数据。

        Returns:
            Updated self.response

        """
        headers = self.headers
        response = dict()
        flag = False

        logging.getLogger('core').info(f'query:{self._request_word}')

        while flag is False:
            page = self._current
            try:
                self.payloads['offset'] = self._offsets[page]
            except IndexError:
                break
            except TypeError:
                break

            url = self.massive_api_parse(self.api, self.payloads)
            url_response = requests.get(url,
                                        headers=headers,
                                        timeout=60  # Avoid blocking
                                        )
            if url_response.ok:
                response[f"page:{page}"] = url_response.json()['row_data']
                logging.getLogger('core').info(f"page {page} is completed.")
                self._current += 1
            else:
                logging.getLogger('core').error(url_response.json())
                raise requests.exceptions.RequestException(url_response.json())

            if len(response[f"page:{page}"]) < self.payloads['pageSize'] and len(self._offsets) <= self._current - 1:
                flag = True

        self.response = response

    def url_requester(self):
        """Direct access to MassIVE database API. 直接访问 MassIVE 数据库 API 的方法。"""
        raise NotImplementedError

    def get_payloads_on_web(self):
        """Access the URL and get the list of parameters of the URL through redirection.
           访问URL并通过重定向获得URL的参数列表。
        """
        url_response = self.url_requester()

        if url_response.ok:
            url_response_json = url_response.json()
            self._total = int(url_response_json['total_rows'])
            self._offsets = [i for i in range(0, self._total, self.payloads['pageSize'])]
            self._current = 0

        return url_response

    @staticmethod
    def massive_api_parse(base_url, payload):
        url_params = f"{base_url}?" \
                     f"pageSize={payload['pageSize']}&" \
                     f"offset={payload['offset']}&" \
                     f"query_type={payload['query_type']}&"

        query = ''
        for k in payload['query'].keys():
            query = f"{query}%2C%22{k}%22%3A%22{payload['query'][k]}%22"

        url_query = f"query=%23%7B{query}%7D"

        url = url_params + url_query

        return url


class MassIVEProjectRetriever(GenericMassIVERetriever):
    """Used to collect project information in MassIVE database. 用于收集 MassIVE 数据库中的项目信息。"""

    def __init__(self):
        super().__init__()
        self.api = 'http://massive.ucsd.edu/ProteoSAFe/proxi/v0.1/datasets'
        self.example = 'PXD014834'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Allows searching for items assigned a number by ProteomeXchange or MassIVE.
        允许搜索被 ProteomeXchange 或 MassIVE 分配编号的项目。

        Args:
            request_word: keyword 请求关键词
        """
        if re.search("pxd", request_word, re.I):
            if re.search("pxd\\d{6}", request_word, re.I).group() == request_word:
                self._request_word = request_word

        elif re.search("msv", request_word, re.I):
            if re.search("MSV\\d{9}", request_word, re.I).group() == request_word:
                self._request_word = request_word

        else:
            raise TypeError('Input error! 输入错误！')

    def retrieve(self):
        """Used to request project data from MassIVE database. 用于向 MassIVE 数据库请求项目数据。

        Returns:
            None

        Updates:
             self.response

        """
        url_response = self.url_requester()
        # logging.getLogger('core').debug('query:{}'.format(self._request_word))

        if url_response.ok:
            url_response = url_response.json()

        else:
            logging.error(f'No correct return transmission received! 未收到正确回传！[{self._request_word}]')
            url_response = dict()

        time.sleep(.3)
        self.response = url_response
        return self.response

    def url_requester(self):
        """API for direct access to MassIVE database project data. 用于直接访问 MassIVE 数据库项目数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}

        self.url = "/".join((self.api, self._request_word))

        url_response = requests.get(self.url, headers=self.headers, params=self.payloads,
                                    timeout=120  # Avoid blocking
                                    )

        return url_response

    def get_payloads_on_web(self):
        """Occupy a seat 占位"""
        raise NotImplementedError("Don\'t need that in this class. 该类不需要这个！")


class MassIVEPeptideRetriever(GenericMassIVERetriever):
    """Used to collect peptide information in MassIVE database. 用于收集 MassIVE 数据库中的肽段信息。"""

    def __init__(self):
        super().__init__()
        self.api = "http://massive.ucsd.edu/ProteoSAFe/QueryPROXI"
        self.example = 'DDDIAALVVDNGSGMCKAGFAGDDAPR'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word

        """
        check = re.search("\\D*", request_word, re.I)
        if check and check.group() == request_word:
            self._request_word = request_word
        else:
            raise TypeError('Input error! 输入错误！')

    def url_requester(self):
        """API for direct access to MassIVE database peptide data. 用于直接访问 MassIVE 数据库肽段数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {'pageSize': 200,
                             'offset': 0,
                             'query': {'peptide': self._request_word},
                             'query_type': 'psm'
                             }

        payloads = self.payloads.copy()
        payloads['pageSize'] = 1
        self.url = self.massive_api_parse(self.api, payloads)

        url_response = requests.get(self.url, headers=self.headers)
        return url_response


class MassIVEProteinRetriever(GenericMassIVERetriever):
    """Used to collect protein information in MassIVE database. 用于收集 MassIVE 数据库中的蛋白质信息。"""

    def __init__(self):
        super().__init__()
        self.api = "http://massive.ucsd.edu/ProteoSAFe/QueryPROXI"
        self.example = 'P60709'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Only UniProt numbers are allowed.
        仅允许UniProt编号

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word
        """
        if re.match('[OPQ]\\d[A-Z\\d]{3}\\d|[A-NR-Z]\\d([A-Z][A-Z\\d]{2}\\d){1,2}',
                    request_word).group() == request_word:
            self._request_word = request_word
        else:
            raise TypeError('Input error! 输入错误！')

    def url_requester(self):
        """API for direct access to MassIVE database protein data. 用于直接访问 MassIVE 数据库蛋白质数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {'pageSize': 200,
                             'offset': 0,
                             'query': {'protein': f"*|{self._request_word}|*"},
                             'query_type': 'protein'
                             }

        payloads = self.payloads.copy()
        payloads['pageSize'] = 1
        self.url = self.massive_api_parse(self.api, payloads)

        url_response = requests.get(self.url, headers=self.headers)
        return url_response


class MassIVEPro2ProjectRetriever(GenericMassIVERetriever):
    """Used to retriever project information by protein accession in the MassIVE database.
       用于在 MassIVE 数据库中按蛋白质编号检索项目信息。
    """

    def __init__(self):
        super().__init__()
        self.api = "http://massive.ucsd.edu/ProteoSAFe/QueryPROXI"
        self.example = 'P60709'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word

        """
        if re.match('[OPQ]\\d[A-Z\\d]{3}\\d|[A-NR-Z]\\d([A-Z][A-Z\\d]{2}\\d){1,2}',
                    request_word).group() == request_word:
            self._request_word = request_word
        else:
            raise TypeError('Input error! 输入错误！')

    def url_requester(self):
        """API for direct access to MassIVE database project data. 用于直接访问 MassIVE 数据库项目数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {'pageSize': 200,
                             'offset': 0,
                             'query': {'protein': f"*|{self._request_word}|*"},
                             'query_type': 'dataset'
                             }

        payloads = self.payloads.copy()
        payloads['pageSize'] = 1
        self.url = self.massive_api_parse(self.api, payloads)

        url_response = requests.get(self.url, headers=self.headers)
        return url_response


class MassIVEPep2ProjectRetriever(GenericMassIVERetriever):
    """Used to retriever peptide sequence by protein accession in the MassIVE database.
       用于在 MassIVE 数据库中按肽段序列检索项目信息。
    """

    def __init__(self):
        super().__init__()
        self.api = "http://massive.ucsd.edu/ProteoSAFe/QueryPROXI"
        self.example = 'DDDIAALVVDNGSGMCKAGFAGDDAPR'

    @property
    def request_word(self):
        return self._request_word

    @request_word.setter
    def request_word(self, request_word):
        """Used to set request keyword. 用于设置请求关键词。

        Args:
            request_word: keyword 请求关键词

        Returns:
            None

        Updates:
            self._request_word

        """
        check = re.search("\\D*", request_word, re.I)
        if check and check.group() == request_word:
            self._request_word = request_word
        else:
            raise TypeError('Input error! 输入错误！')

    def url_requester(self):
        """API for direct access to MassIVE database project data. 用于直接访问 MassIVE 数据库项目数据的API。

        Returns:
            Accessing the API returns a requests object. 访问API返回一个requests对象。

        """
        if self.headers is None:
            self.headers = {'Accept': 'application/json'}
        if self.payloads is None:
            self.payloads = {'pageSize': 200,
                             'offset': 0,
                             'query': {'peptide': self._request_word},
                             'query_type': 'dataset'
                             }

        payloads = self.payloads.copy()
        payloads['pageSize'] = 1
        self.url = self.massive_api_parse(self.api, payloads)

        url_response = requests.get(self.url, headers=self.headers)
        return url_response
